from __future__ import annotations

import logging
from collections.abc import MutableMapping
from pathlib import Path
from typing import Any, NamedTuple

import keyring
import keyring.backends.fail
import keyring.backends.null
from kraken.common import EnvironmentType, http
from kraken.common.http import ReadTimeout

logger = logging.getLogger(__name__)
DEFAULT_CONFIG_PATH = Path("~/.config/krakenw/config.toml").expanduser()


def type_fqn(x: type[Any] | object) -> str:
    if not isinstance(x, type):
        x = type(x)
    return f"{x.__module__}.{x.__name__}"


class AuthModel:
    """Provides an interface to store credentials safely in the system keyring. If keyring backend is available,
    the credentials are stored in the config file instead."""

    class CredentialsWithHost(NamedTuple):
        host: str
        username: str
        password: str

    class Credentials(NamedTuple):
        username: str
        password: str

    class CredentialCheck(NamedTuple):
        curl_command: str
        auth_check_result: bool
        raw_result: str
        hint: str

    def __init__(self, config: MutableMapping[str, Any], path: Path, use_keyring_if_available: bool) -> None:
        self._config = config
        self._path = path
        self._has_keyring = use_keyring_if_available and not isinstance(
            keyring.get_keyring(), (keyring.backends.fail.Keyring, keyring.backends.null.Keyring)
        )

    def get_credentials(self, host: str) -> Credentials | None:
        auth = self._config.get("auth", {})
        if host not in auth:
            return None
        username = auth[host].get("username")
        if not username:
            return None
        password = auth[host].get("password")
        if password is not None:
            return self.Credentials(username, password)
        if self._has_keyring:
            password = keyring.get_credential(host, username)
            if not password:
                return None
            return self.Credentials(username, password.password)
        return None

    def set_credentials(self, host: str, username: str, password: str) -> None:
        auth = self._config.setdefault("auth", {})
        auth[host] = {"username": username}
        if not self._has_keyring:
            auth[host]["password"] = password
            logger.warning(
                "no keyring backend available (%s), password will be stored in plain text",
                type_fqn(keyring.get_keyring()),
            )
            if isinstance(keyring.get_keyring(), keyring.backends.null.Keyring):
                logger.warning(
                    "it looks like you may have disabled keyring globally. consider re-enabling it by running `python "
                    "-m keyring diagnose` to find the config file and either removing the file entirely or removing "
                    "the `keyring.backends.null.Keyring` entry from the `default-backend` key in the `[backend]` "
                    "section. (the file is usually located at `~/.local/share/python_keyring/keyringrc.cfg` "
                    "or `~/.config/python_keyring/keyringrc.cfg`)"
                )
            logger.info("saving username and password for %s in %s", host, self._path)
        else:
            logger.debug("keyring backend available (%s)", type_fqn(keyring.get_keyring()))
            logger.info("saving username for %s in %s", host, self._path)
            logger.info("saving password for %s in keyring", host)
            keyring.set_password(host, username, password)

    def delete_credentials(self, host: str) -> None:
        auth = self._config.get("auth")
        if auth and host in auth:
            username = auth[host].get("username")
            logger.info("deleting username for %s from %s", host, self._path)
            del auth[host]
            if username and self._has_keyring:
                logger.info("deleting password for %s from keyring", host)
                try:
                    keyring.delete_password(host, username)
                except keyring.errors.PasswordDeleteError:
                    logger.warning("item in keychain not found")
        else:
            logger.warning("no credentials entry for %s", host)

    def list_credentials(self) -> list[CredentialsWithHost]:
        result = []
        for host in self._config.get("auth", {}):
            credentials = self.get_credentials(host)
            if credentials:
                result.append(self.CredentialsWithHost(host, *credentials))
        return result

    def check_credential(self, host: str, username: str, password: str) -> CredentialCheck | None:
        if "jfrog" in host:
            # Allow the user to override the url that will be used by setting
            # the `auth_check_url_suffix` in their krakenw/config.toml file PER HOST
            url_suffix = (
                self._config.get("auth", {})
                .get(host, {})
                .get("auth_check_url_suffix", "artifactory/api/pypi/python-all/simple/flask/")
            )
            url = f"https://{host}/{url_suffix}"
            curl_command = f"curl --user '{username}:{password}' {url}"

            # Get the result
            try:
                result = http.get(url, auth=(username, password), timeout=10)
            except ReadTimeout:
                logger.warning("HTTP Timeout when testing credentials")
                return None

            # Build hints
            hints = []

            if result.status_code == 401:
                if "Props authentication" in result.text:
                    hints.append("You may have used an API token rather than an identity token.")
                    hints.append("Your username and/or token may be incorrect.")
                elif "Token principal mismatch" in result.text:
                    hints.append("Your username and/or token may be incorrect.")
                elif "Bad credentials" in result.text:
                    hints.append("Your credentials are invalid")
                else:
                    hints.append("Credential check resulted in unknown 401 unauthorised error")
            elif result.status_code in (404, 302):
                hints.append(
                    f"""Authentication URL incorrect (HTTP response {result.status_code}).
Please check host.auth_check_url_suffix value in {self._path}"""
                )

            return self.CredentialCheck(curl_command, result.status_code == 200, result.text, " ".join(hints))

        if "gitlab." in host:
            # Allow the user to override the url that will be used by setting
            # the `auth_check_url_suffix` in their krakenw/config.toml file PER HOST
            url_suffix = self._config.get("auth", {}).get(host, {}).get("auth_check_url_suffix", "api/v4/projects")
            url = f"https://{host}/{url_suffix}"

            # Get the result
            try:
                result = http.get(url, params={"access_token": password}, timeout=10)
            except ReadTimeout:
                logger.warning("HTTP Timeout when testing credentials")
                return None
            curl_command = "curl " + str(result.url)

            # Build hints
            hints = []
            if result.status_code == 401:
                hints.append("Your credentials are invalid")

            return self.CredentialCheck(curl_command, result.status_code == 200, result.text, " ".join(hints))

        return None


class ConfigModel:
    def __init__(self, config: MutableMapping[str, Any], path: Path) -> None:
        self._config = config
        self._path = path

    def set_default_installer(self, env_type: EnvironmentType) -> None:
        self._config["default_installer"] = env_type.name
        logger.info("saving default_installer=%r in %s", env_type.name, self._path)

    def get_default_installer(self) -> EnvironmentType:
        default = EnvironmentType.VENV
        if value := self._config.get("default_installer"):
            try:
                return EnvironmentType[value]
            except KeyError:
                logger.warning("invalid default_installer: %r, falling back to %r", value, default.name)
        return default



from __future__ import annotations

import abc
import dataclasses
import datetime
import json
import logging
import subprocess
from collections.abc import Sequence
from pathlib import Path
from typing import Any, NoReturn

from kraken.common import EnvironmentType, NotSet, RequirementSpec, datetime_to_iso8601, iso8601_to_datetime

from ._lockfile import Distribution

logger = logging.getLogger(__name__)

KRAKEN_MAIN_IMPORT_SNIPPET = """
try:
    from kraken.core.cli.main import main  # >= 0.9.0
except ImportError:
    from kraken.cli.main import main  # < 0.9.0
""".strip()


class BuildEnv(abc.ABC):
    """Interface for the build environment."""

    @abc.abstractmethod
    def get_type(self) -> EnvironmentType:
        """Return the type of build environment that this is."""

    @abc.abstractmethod
    def get_path(self) -> Path:
        """Return the path to the build environment."""

    @abc.abstractmethod
    def get_installed_distributions(self) -> list[Distribution]:
        """Return the distributions that are currently installed in the environment."""

    @abc.abstractmethod
    def build(self, requirements: RequirementSpec, transitive: bool) -> None:
        """Build the environment from the given requirement spec."""

    @abc.abstractmethod
    def dispatch_to_kraken_cli(self, argv: list[str]) -> NoReturn:
        """Dispatch the kraken cli command in *argv* to the build environment.

        :param argv: The arguments to pass to the kraken cli (without the "kraken" command name itself)."""


@dataclasses.dataclass(frozen=True)
class BuildEnvMetadata:
    created_at: datetime.datetime
    environment_type: EnvironmentType
    requirements_hash: str
    hash_algorithm: str

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> BuildEnvMetadata:
        return cls(
            created_at=iso8601_to_datetime(data["created_at"]),
            environment_type=EnvironmentType[data["environment_type"]],
            requirements_hash=data["requirements_hash"],
            hash_algorithm=data["hash_algorithm"],
        )

    def to_json(self) -> dict[str, Any]:
        return {
            "created_at": datetime_to_iso8601(self.created_at),
            "environment_type": self.environment_type.name,
            "requirements_hash": self.requirements_hash,
            "hash_algorithm": self.hash_algorithm,
        }


@dataclasses.dataclass
class BuildEnvMetadataStore:
    path: Path

    def __post_init__(self) -> None:
        self._metadata: BuildEnvMetadata | None | NotSet = NotSet.Value

    def get(self) -> BuildEnvMetadata | None:
        if self._metadata is NotSet.Value:
            if self.path.is_file():
                self._metadata = BuildEnvMetadata.from_json(json.loads(self.path.read_text()))
            else:
                self._metadata = None
        return self._metadata

    def set(self, metadata: BuildEnvMetadata) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self.path.write_text(json.dumps(metadata.to_json()))
        self._metadata = metadata


class BuildEnvError(Exception):
    """
    An error occurred while building the environment.
    """


def general_get_installed_distributions(kraken_command_prefix: Sequence[str]) -> list[Distribution]:
    command = [*kraken_command_prefix, "query", "env"]
    output = subprocess.check_output(command).decode()
    return [Distribution(x["name"], x["version"], x["requirements"], x["extras"]) for x in json.loads(output)]



from pathlib import Path
from typing import TYPE_CHECKING, Any

from kraken.common import EnvironmentType, RequirementSpec

from ._buildenv_venv import VenvBuildEnv as _VenvBuildEnv

if TYPE_CHECKING:

    def find_uv_bin() -> str: ...

else:
    from uv.__main__ import find_uv_bin


class UvBuildEnv(_VenvBuildEnv):
    """Implements a build environment managed by the `uv` package manager.

    This has a lot in common with the `venv` build environment, so we just inherit from it and override the relevant
    parts.
    """

    INSTALLER_NAME = "uv"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._uv_bin = find_uv_bin()

    def _get_create_venv_command(self, python_bin: Path, path: Path) -> list[str]:
        return [self._uv_bin, "venv", str(path)]

    def _get_install_command(self, venv_dir: Path, requirements: RequirementSpec, env: dict[str, str]) -> list[str]:
        env["VIRTUAL_ENV"] = str(venv_dir)
        return [self._uv_bin, "pip", "install", *requirements.to_args(base_dir=self._project_root)]

    def get_type(self) -> EnvironmentType:
        return EnvironmentType.UV



from __future__ import annotations

import dataclasses
import re
from pathlib import Path
from typing import Any

from kraken.common import LocalRequirement, PipRequirement, RequirementSpec
from packaging.requirements import Requirement as ParsedRequirement


@dataclasses.dataclass(frozen=True)
class Distribution:
    name: str
    version: str
    requirements: list[str]
    extras: set[str]


@dataclasses.dataclass
class Lockfile:
    requirements: RequirementSpec
    pinned: dict[str, str]

    @staticmethod
    def from_path(path: Path) -> Lockfile:
        import tomli

        with path.open("rb") as fp:
            return Lockfile.from_json(tomli.load(fp))

    def write_to(self, path: Path) -> None:
        import tomli_w

        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("wb") as fp:
            tomli_w.dump(self.to_json(), fp)

    @staticmethod
    def from_json(data: dict[str, Any]) -> Lockfile:
        return Lockfile(
            requirements=RequirementSpec.from_json(data["requirements"]),
            pinned=data["pinned"],
        )

    def to_json(self) -> dict[str, Any]:
        return {
            "requirements": self.requirements.to_json(),
            "pinned": self.pinned,
        }

    def to_pinned_requirement_spec(self) -> RequirementSpec:
        """Converts the pinned versions in the lock file to a :class:`RequirementSpec` with the pinned requirements."""

        requirements = RequirementSpec(
            requirements=(),
            index_url=self.requirements.index_url,
            extra_index_urls=self.requirements.extra_index_urls[:],
            pythonpath=self.requirements.pythonpath[:],
            interpreter_constraint=self.requirements.interpreter_constraint,
        )

        # Make sure that local requirements keep being installed from the local source.
        local_requirements = {
            dep.name: dep for dep in self.requirements.requirements if isinstance(dep, LocalRequirement)
        }
        requirements = requirements.with_requirements(local_requirements.values())

        # Add all non-local requirements with exact version numbers.
        requirements = requirements.with_requirements(
            f"{key}=={value}" for key, value in sorted(self.pinned.items()) if key not in local_requirements
        )

        return requirements


def normalize_package_name(name: str) -> str:
    # PEP503/PEP426 normalized package name.
    return re.sub(r"[-_.]+", "-", name).lower()


def calculate_lockfile(
    requirements: RequirementSpec,
    distributions: list[Distribution],
) -> tuple[Lockfile, set[str]]:
    """Calculate the lockfile of the environment.

    :param requirements: The requirements that were used to install the environment. These requirements
        will be embedded as part of the returned lockfile.
    :return: (lockfile, extra_distributions)
    """

    # Contains the versions we pinned.
    pinned: dict[str, str] = {}

    # Normalize distribution names.
    dists = {normalize_package_name(dist.name): dist for dist in distributions}

    # Convert our internal requirements representation to parsed requirements. Local requirements
    # are treated without extras.
    requirements_stack = [
        ParsedRequirement(str(req) if isinstance(req, PipRequirement) else req.name)
        for req in requirements.requirements
    ]

    while requirements_stack:
        package_req = requirements_stack.pop(0)
        package_name = normalize_package_name(package_req.name)

        if package_name in pinned:
            # Already collected it.
            # TODO (@NiklasRosenstein): Maybe this req has extras we haven't considered yer?
            # NOTE (@NiklasRosenstein): We may be missing the package because it's a requirement that is only
            #       installed under certain conditions (e.g. markers/extras).
            continue

        # Pin the package version.
        dist = dists[package_name]
        pinned[package_name] = dist.version

        # Filter the requirements of the distribution down to the ones required according to markers and the
        # current package requirement's extras.
        for req in map(ParsedRequirement, dist.requirements):
            if not req.marker or any(req.marker.evaluate({"extra": extra}) for extra in package_req.extras):
                requirements_stack.append(req)

    extra_distributions = dists.keys() - pinned.keys()
    pinned = {k: v for k, v in sorted(pinned.items(), key=lambda t: t[0].lower())}
    return Lockfile(requirements, pinned), extra_distributions



from __future__ import annotations

import logging
import os
import shlex
import subprocess
import sys
from contextlib import ExitStack
from pathlib import Path
from typing import ClassVar, Literal, NoReturn

from kraken.common import EnvironmentType, RequirementSpec, findpython, safe_rmpath
from kraken.common.pyenv import VirtualEnvInfo

from ._buildenv import KRAKEN_MAIN_IMPORT_SNIPPET, BuildEnv, BuildEnvError, general_get_installed_distributions
from ._lockfile import Distribution

logger = logging.getLogger(__name__)


def find_python_interpreter(constraint: str) -> str:
    """
    Finds a Python interpreter that matches the given constraint. We rely on the order of candidates returned by
    #findpython.get_candidates() and return the first matching Python version.
    """

    interpreters = findpython.evaluate_candidates(findpython.get_candidates(), findpython.InterpreterVersionCache())
    for interpreter in interpreters:
        if findpython.match_version_constraint(constraint, interpreter["version"]):
            return interpreter["path"]

    raise RuntimeError(f"Could not find a Python interpreter that matches the constraint {constraint!r}.")


class VenvBuildEnv(BuildEnv):
    """
    Installs the Kraken build environment into a Python virtual environment.
    """

    INSTALLER_NAME: ClassVar[str] = "Pip"

    def __init__(self, project_root: Path, path: Path, incremental: bool = False, show_pip_logs: bool = False) -> None:
        """
        Args:
            project_root: Path for resolving relative local requirements.
            path: Path where the virtual env will be created.
            incremental: Whether install operations can re-use an existing state of the virtual environment.
            show_pip_logs: Keep Pip logs attached to the terminal. Otherwise they're piped to a file.
        """

        self._project_root = project_root
        self._path = path
        self._venv = VirtualEnvInfo(self._path)
        self._incremental = incremental
        self._show_pip_logs = show_pip_logs

    def _run_command(
        self,
        command: list[str],
        operation_name: str,
        log_file: Path | None,
        mode: Literal["a", "w"] = "w",
        env: dict[str, str] | None = None,
    ) -> None:
        if log_file:
            log_file.parent.mkdir(parents=True, exist_ok=True)

        offset: int | None = None
        exc: Exception | None = None

        with ExitStack() as stack:
            if log_file is not None:
                fp = stack.enter_context(log_file.open(mode))
                offset = fp.tell()
            else:
                fp = None
                offset = 0
            try:
                subprocess.check_call(command, stdout=fp, stderr=fp, env=env)
                return
            except (subprocess.CalledProcessError, FileNotFoundError) as e:
                exc = e

        assert exc is not None
        exit_code = exc.returncode if isinstance(exc, subprocess.CalledProcessError) else -1
        command_str = "$ " + " ".join(map(shlex.quote, command))

        if log_file:
            assert offset is not None
            with log_file.open() as fp:
                fp.seek(offset)
                logger.error(
                    "'%s' failed (exit code: %d, command: %s). Output:\n\n%s",
                    operation_name,
                    exit_code,
                    command_str,
                    fp.read(),
                )
        else:
            logger.error(
                "'%s' failed (exit code: %d, command: %s). Check the output above for more information.",
                operation_name,
                exc.returncode if isinstance(exc, subprocess.CalledProcessError) else -1,
                "$ " + " ".join(map(shlex.quote, command)),
            )

        raise BuildEnvError(f"The command {command} failed.") from exc

    def _get_create_venv_command(self, python_bin: Path, path: Path) -> list[str]:
        """Returns the command to create a virtual environment."""
        return [str(python_bin), "-m", "venv", str(path), "--upgrade-deps"]

    def _get_install_command(self, venv_dir: Path, requirements: RequirementSpec, env: dict[str, str]) -> list[str]:
        """Return sthe command to install the given requirements into the virtual environment."""
        python_bin = venv_dir / "bin" / "python"
        command = [
            str(python_bin),
            "-m",
            "pip",
            "install",
            "--disable-pip-version-check",
            "--no-python-version-warning",
            "--no-input",
        ]
        # Must enable transitive resolution because lock files are not currently cross platform (see kraken-wrapper#2).
        # if not transitive:
        #     command += ["--no-deps"]
        command += requirements.to_args(base_dir=self._project_root)
        return command

    def _install_pythonpath(self, venv_dir: Path, pythonpath: list[str]) -> None:
        """Install the given pythonpath into the virtual environment."""
        python_bin = venv_dir / "bin" / "python"
        command = [str(python_bin), "-c", "from sysconfig import get_path; print(get_path('purelib'))"]
        site_packages = Path(subprocess.check_output(command).decode().strip())
        pth_file = site_packages / "krakenw.pth"
        if pythonpath:
            logger.debug("Writing .pth file at %s", pth_file)
            pth_file.write_text("\n".join(str(Path(path).absolute()) for path in pythonpath))
        elif pth_file.is_file():
            logger.debug("Removing .pth file at %s", pth_file)
            pth_file.unlink()

    # BuildEnv

    def get_path(self) -> Path:
        return self._path

    def get_type(self) -> EnvironmentType:
        return EnvironmentType.VENV

    def get_installed_distributions(self) -> list[Distribution]:
        python = self._venv.get_bin("python")
        return general_get_installed_distributions([str(python), "-c", f"{KRAKEN_MAIN_IMPORT_SNIPPET}\nmain()"])

    def build(self, requirements: RequirementSpec, transitive: bool) -> None:
        if self._show_pip_logs:
            create_log: Path | None = self._path.with_name(self._path.name + ".log") / "create.txt"
            install_log: Path | None = self._path.with_name(self._path.name + ".log") / "install.txt"
        else:
            create_log = install_log = None

        if not self._incremental and self._path.exists():
            logger.debug("Removing existing virtual environment at %s", self._path)
            safe_rmpath(self._path)

        python_bin = str(self._venv.get_bin("python"))
        success_flag = self._path / ".success.flag"

        # If a virtual environment already exists, we should ensure that it matches the given interpreter constraint.
        if os.path.isfile(python_bin) and success_flag.is_file():
            try:
                current_python_version = findpython.get_python_interpreter_version(python_bin)
            except (subprocess.CalledProcessError, RuntimeError) as e:
                logger.warning("Could not determine the version of the current Python build environment: %s", e)
                logger.info("Destroying existing environment at %s", self._path)
                safe_rmpath(self._path)
            else:
                if requirements.interpreter_constraint and not findpython.match_version_constraint(
                    requirements.interpreter_constraint, current_python_version
                ):
                    logger.info(
                        "Existing Python interpreter at %s does not match constraint %s because its Python version "
                        "is %s. The environment will be recreated with the correct interpreter.",
                        python_bin,
                        requirements.interpreter_constraint,
                        current_python_version,
                    )
                    safe_rmpath(self._path)

        elif self._venv.exists() and not success_flag.is_file():
            logger.warning("Your virtual build environment appears to be corrupt. It will be recreated. This happens")
            logger.warning("by pressing Ctrl+C during its installation, or if you've recently upgraded kraken-wrapper.")
            safe_rmpath(self._path)

        if not self._path.exists():
            # Find a Python interpreter that matches the given interpreter constraint.
            if requirements.interpreter_constraint is not None:
                logger.info("Using Python interpreter constraint: %s", requirements.interpreter_constraint)
                python_origin_bin = find_python_interpreter(requirements.interpreter_constraint)
                logger.info("Using Python interpreter at %s", python_origin_bin)
            else:
                logger.info(
                    "No interpreter constraint specified, using current Python interpreter (%s)", sys.executable
                )
                python_origin_bin = sys.executable

            command = self._get_create_venv_command(Path(python_origin_bin), self._path)
            logger.info("Creating virtual environment at %s", os.path.relpath(self._path))
            self._run_command(command, operation_name="Create virtual environment", log_file=create_log)
            success_flag.touch()

        else:
            logger.info("Reusing virtual environment at %s", self._path)

        # Install requirements.
        env = os.environ.copy()
        command = self._get_install_command(self._path, requirements, env)
        logger.info("Installing dependencies.")
        logger.debug("Installing into build environment with %s: %s", self.INSTALLER_NAME, " ".join(command))
        self._run_command(command, operation_name="Install dependencies", log_file=install_log, env=env)

        # Make sure the pythonpath from the requirements is encoded into the enviroment.
        self._install_pythonpath(self._path, list(requirements.pythonpath))

    def dispatch_to_kraken_cli(self, argv: list[str]) -> NoReturn:
        python = self._venv.get_bin("python")
        command = [str(python), "-c", f"{KRAKEN_MAIN_IMPORT_SNIPPET}\nmain()", *argv]
        env = os.environ.copy()
        self.get_type().set(env)
        env["PATH"] = str(self._venv.get_bin_directory()) + os.pathsep + env.get("PATH", "")
        sys.exit(subprocess.call(command, env=env))



from __future__ import annotations

import argparse
import dataclasses
import os

from kraken.common import EnvironmentType


@dataclasses.dataclass(frozen=True)
class EnvOptions:
    status: bool
    upgrade: bool
    reinstall: bool
    uninstall: bool
    use: EnvironmentType | None
    incremental: bool
    show_install_logs: bool
    no_keyring: bool

    @staticmethod
    def add_to_parser(parser: argparse.ArgumentParser) -> None:
        group = parser.add_argument_group("build environment")
        group.add_argument(
            "--use",
            choices=[v.name for v in EnvironmentType if v.is_wrapped()],
            default=os.getenv("KRAKENW_USE"),
            help="use the specified environment type. If the environment type changes it will trigger a reinstall.\n"
            "Defaults to the value of the KRAKENW_USE environment variable. If that variable is unset, and\nif a build "
            "environment already exists, that environment's type will be used. The default\nenvironment type that is "
            "used for new environments is VENV. [env: KRAKENW_USE=...]",
        )
        group.add_argument(
            "--status",
            action="store_true",
            help="print the status of the build environment and exit",
        )
        group.add_argument(
            "--upgrade",
            action="store_true",
            help="reinstall the build environment from the original requirements",
        )
        group.add_argument(
            "--reinstall",
            action="store_true",
            default=os.getenv("KRAKENW_REINSTALL") == "1",
            help="reinstall the build environment from the lock file [env: KRAKENW_REINSTALL=1]",
        )
        group.add_argument(
            "--uninstall",
            action="store_true",
            help="uninstall the build environment",
        )
        group.add_argument(
            "--incremental",
            action="store_true",
            default=os.getenv("KRAKENW_INCREMENTAL") == "1",
            help="re-use an existing build environment. Improves installation time after an update to the buildscript\n"
            "dependencies, but does not upgrade all packages to latest. [env: KRAKENW_INCREMENTAL=1]",
        )
        group.add_argument(
            "--show-install-logs",
            action="store_true",
            default=os.getenv("KARKENW_SHOW_INSTALL_LOGS") == "1",
            help="show Pip install logs instead of piping them to the build/.venv.log/ directory.\n"
            "[env: KARKENW_SHOW_INSTALL_LOGS=1]",
        )

        group = parser.add_argument_group("authentication")
        group.add_argument(
            "--no-keyring",
            action="store_true",
            default=os.getenv("KRAKENW_NO_KEYRING") == "1",
            help="disable the use of the keyring package for loading and storing credentials. "
            "[env: KRAKENW_NO_KEYRING=1]",
        )

    @classmethod
    def collect(cls, args: argparse.Namespace) -> EnvOptions:
        return cls(
            status=args.status,
            upgrade=args.upgrade,
            reinstall=args.reinstall,
            uninstall=args.uninstall,
            use=EnvironmentType[args.use] if args.use else None,
            incremental=args.incremental,
            show_install_logs=args.show_install_logs,
            no_keyring=args.no_keyring,
        )

    def any(self) -> bool:
        return bool(self.status or self.upgrade or self.reinstall or self.uninstall or self.use)


@dataclasses.dataclass(frozen=True)
class AuthOptions:
    host: str
    username: str | None
    password: str | None
    password_stdin: bool
    remove: bool
    list: bool
    no_check: bool
    no_mask: bool
    verbose: bool

    @staticmethod
    def add_to_parser(parser: argparse.ArgumentParser) -> None:
        parser.add_argument("host", nargs="?", help="the host to add the credentials for")
        parser.add_argument(
            "-u",
            "--username",
            help="the username to use when accessing resources on the given host",
        )
        parser.add_argument(
            "-p",
            "--password",
            help="the password to use when accessing resources on the given host (use --password-stdin when possible)",
        )
        parser.add_argument("--password-stdin", action="store_true", help="read the password from stdin")
        parser.add_argument(
            "-r",
            "--remove",
            action="store_true",
            help="remove credentials for the given host",
        )
        parser.add_argument(
            "-l",
            "--list",
            action="store_true",
            help="list configured credentials for the given host",
        )
        parser.add_argument(
            "-s",
            "--no-check",
            action="store_true",
            help="skip checking of auth credentials",
        )
        parser.add_argument(
            "--no-mask",
            action="store_true",
            help="do not mask credentials",
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            default=False,
            help="show curl queries to use when authenicating hosts",
        )

    @classmethod
    def collect(cls, args: argparse.Namespace) -> AuthOptions:
        return cls(
            host=args.host,
            username=args.username,
            password=args.password,
            password_stdin=args.password_stdin,
            remove=args.remove,
            list=args.list,
            no_check=args.no_check,
            no_mask=args.no_mask,
            verbose=args.verbose,
        )



from __future__ import annotations

import datetime
import hashlib
import logging
import platform
from pathlib import Path
from urllib.parse import quote, urlparse, urlunparse

from kraken.common import EnvironmentType, RequirementSpec, not_none, safe_rmpath

from ._buildenv import BuildEnv, BuildEnvMetadata, BuildEnvMetadataStore
from ._buildenv_uv import UvBuildEnv
from ._buildenv_venv import VenvBuildEnv
from ._config import AuthModel
from ._lockfile import Lockfile

logger = logging.getLogger(__name__)


class BuildEnvManager:
    def __init__(
        self,
        project_root: Path,
        path: Path,
        auth: AuthModel,
        default_type: EnvironmentType = EnvironmentType.VENV,
        default_hash_algorithm: str = "sha256",
        incremental: bool = False,
        show_install_logs: bool = False,
    ) -> None:
        """
        Args:
            project_root: Path for resolving relative local requirements.
            path: Path to the directory that contains the build environment (virtual env).
        """

        assert (
            default_hash_algorithm in hashlib.algorithms_available
        ), f"hash algoritm {default_hash_algorithm!r} is not available"

        self._project_root = project_root
        self._path = path
        self._auth = auth
        self._metadata_store = BuildEnvMetadataStore(path.parent / (path.name + ".meta"))
        self._default_type = default_type
        self._default_hash_algorithm = default_hash_algorithm
        self._incremental = incremental
        self._show_install_logs = show_install_logs

    def _inject_auth(self, url: str) -> str:
        parsed_url = urlparse(url)
        credentials = self._auth.get_credentials(parsed_url.netloc)
        if credentials is None:
            return url

        logger.info('Injecting username and password into index url "%s"', url)
        domain = parsed_url.netloc.rpartition("@")[-1]
        parsed_url = parsed_url._replace(netloc=f"{quote(credentials.username)}:{quote(credentials.password)}@{domain}")
        url = urlunparse(parsed_url)
        return url

    def exists(self) -> bool:
        if self._metadata_store.get() is None:
            return False  # If we don't have metadata, we assume the environment does not exist.
        return self.get_environment(None).get_path().exists()

    def remove(self) -> None:
        safe_rmpath(self._metadata_store.path)
        safe_rmpath(self.get_environment(None).get_path())

    def install(
        self,
        requirements: RequirementSpec,
        env_type: EnvironmentType | None = None,
        transitive: bool = True,
        allow_incremental: bool = True,
    ) -> None:
        """
        Args:
            requirements: The requirements to build the environment with.
            env_type: The environment type to use. If not specified, falls back to the last used or default.
            transitive: If set to `False`, it indicates that the *requirements* are fully resolved and the
                        build environment installer does not need to resolve transitve dependencies.
            allow_incremental: Allow incremental builds if the environment already exists. Set to False if
                               the environment type changes.
        """

        if env_type is None:
            metadata = self._metadata_store.get()
            env_type = metadata.environment_type if metadata else self._default_type

        # Inject credentials into the requirements.
        requirements = RequirementSpec(
            requirements=requirements.requirements,
            index_url=self._inject_auth(requirements.index_url) if requirements.index_url else None,
            extra_index_urls=tuple(self._inject_auth(url) for url in requirements.extra_index_urls),
            interpreter_constraint=requirements.interpreter_constraint,
            pythonpath=requirements.pythonpath,
        )

        env = self.get_environment(env_type, allow_incremental)
        env.build(requirements, transitive)
        hash_algorithm = self.get_hash_algorithm()
        metadata = BuildEnvMetadata(
            datetime.datetime.now(datetime.timezone.utc),
            env.get_type(),
            requirements.to_hash(hash_algorithm),
            hash_algorithm,
        )
        self._metadata_store.set(metadata)

    def get_metadata_file(self) -> Path:
        return self._metadata_store.path

    def get_metadata(self) -> BuildEnvMetadata:
        return not_none(self._metadata_store.get(), "metadata does not exist")

    def get_hash_algorithm(self) -> str:
        metadata = self._metadata_store.get()
        return metadata.hash_algorithm if metadata else self._default_hash_algorithm

    def get_environment(self, env_type: EnvironmentType | None = None, allow_incremental: bool = True) -> BuildEnv:
        if env_type is None:
            metadata = self._metadata_store.get()
            env_type = self._default_type if metadata is None else metadata.environment_type
        return _get_environment_for_type(
            env_type, self._project_root, self._path, self._incremental and allow_incremental, self._show_install_logs
        )

    def set_locked(self, lockfile: Lockfile) -> None:
        metadata = self._metadata_store.get()
        assert metadata is not None
        metadata = BuildEnvMetadata(
            metadata.created_at,
            metadata.environment_type,
            lockfile.to_pinned_requirement_spec().to_hash(metadata.hash_algorithm),
            metadata.hash_algorithm,
        )
        self._metadata_store.set(metadata)


def _get_environment_for_type(
    environment_type: EnvironmentType,
    project_root: Path,
    path: Path,
    incremental: bool,
    show_install_logs: bool,
) -> BuildEnv:
    platform_name = platform.system().lower()
    match environment_type:
        case EnvironmentType.VENV:
            return VenvBuildEnv(
                project_root,
                path,
                incremental=incremental,
                show_pip_logs=show_install_logs,
            )
        case EnvironmentType.UV:
            return UvBuildEnv(
                project_root,
                path,
                incremental=incremental,
                show_pip_logs=show_install_logs,
            )
        case _:
            raise RuntimeError(f"unsupported environment type {environment_type!r} on platform {platform_name!r}")



from __future__ import annotations

import argparse
import getpass
import logging
import os
import shlex
import sys
import time
from pathlib import Path
from textwrap import indent
from typing import NamedTuple, NoReturn

from kraken.common import (
    AsciiTable,
    BuildscriptMetadata,
    EnvironmentType,
    GitAwareProjectFinder,
    LoggingOptions,
    RequirementSpec,
    TomlConfigFile,
    colored,
    datetime_to_iso8601,
    inline_text,
)
from kraken.common.exceptions import exit_on_known_exceptions

from . import __version__
from ._buildenv import BuildEnvError
from ._buildenv_manager import BuildEnvManager
from ._config import DEFAULT_CONFIG_PATH, AuthModel, ConfigModel
from ._lockfile import Lockfile, calculate_lockfile
from ._option_sets import AuthOptions, EnvOptions

BUILDENV_PATH = Path("build/.kraken/venv")
BUILDSCRIPT_FILENAME = ".kraken.py"
BUILD_SUPPORT_DIRECTORY = "build-support"
LOCK_FILENAME = ".kraken.lock"
_FormatterClass = lambda prog: argparse.RawTextHelpFormatter(prog, max_help_position=60, width=120)  # noqa: E731
logger = logging.getLogger(__name__)


def _get_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        "krakenw",
        formatter_class=_FormatterClass,
        description=inline_text(
            f"""
            This is kraken-wrapper v{__version__}.

            {colored("krakenw", attrs=["bold"])} is a thin wrapper around the {colored("kraken", attrs=["bold"])} cli
            that executes builds in an isolated \\
            build environment. This ensures that builds are reproducible (especially when using \\
            lock files).

            To learn more about kraken, visit https://github.com/kraken-build/kraken-core.
            """
        ),
        epilog=inline_text(
            colored(
                "This is kraken-wrapper's help. To show kraken's help instead, run krakenw -- --help",
                "yellow",
                attrs=["bold"],
            )
        ),
    )
    parser.add_argument("-V", "--version", version=__version__, action="version")
    LoggingOptions.add_to_parser(parser, default_verbosity=1)
    EnvOptions.add_to_parser(parser)

    # NOTE (@NiklasRosenstein): If we combine "+" with remainder, we get options passed after the `cmd`
    #       passed directly into `args` without argparse treating it like an option. This is not the case
    #       when using `nargs=1` for `cmd`.
    parser.add_argument("cmd", nargs="*", metavar="cmd", help="{auth,list-pythons,lock} or a kraken command")
    parser.add_argument("args", nargs=argparse.REMAINDER, help="additional arguments")
    return parser


def _get_lock_argument_parser(prog: str) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog,
        formatter_class=_FormatterClass,
        description=inline_text(
            f"""
            Rewrite the lock file ({colored(LOCK_FILENAME, attrs=["bold"])}) from the current build environment.
            """
        ),
    )

    return parser


def lock(prog: str, argv: list[str], manager: BuildEnvManager, project: Project) -> NoReturn:
    parser = _get_lock_argument_parser(prog)
    parser.parse_args(argv)

    if not manager.exists():
        logger.error("cannot lock without a build environment")
        sys.exit(1)

    environment = manager.get_environment()
    distributions = environment.get_installed_distributions()
    lockfile, extra_distributions = calculate_lockfile(project.requirements, distributions)

    if environment.get_type() == EnvironmentType.VENV:
        extra_distributions.discard("pip")  # We'll always have that in a virtual env.

    if extra_distributions:
        logger.warning("Found extra distributions in your Kraken build enviroment: %s", ", ".join(extra_distributions))

    had_lockfile = project.lockfile_path.exists()
    lockfile.write_to(project.lockfile_path)
    manager.set_locked(lockfile)

    logger.info("Lock file %s (%s)", "updated" if had_lockfile else "created", os.path.relpath(project.lockfile_path))
    sys.exit(0)


def _get_auth_argument_parser(prog: str) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog,
        formatter_class=_FormatterClass,
        description=inline_text(
            """
            Configure the credentials to use when accessing PyPI packages from the given host.
            The password will be stored in the system keychain.
            """
        ),
    )
    AuthOptions.add_to_parser(parser)
    return parser


def config_main(prog: str, argv: list[str]) -> NoReturn:
    """`krakenw config` subcommand."""

    config_file = TomlConfigFile(DEFAULT_CONFIG_PATH)
    config = ConfigModel(config_file, DEFAULT_CONFIG_PATH)

    parser = argparse.ArgumentParser(prog=prog)
    parser.add_argument(
        "--installer",
        choices=[x.name for x in EnvironmentType if x.is_wrapped()],
        help="Set the installer to use for the build environment.",
    )
    args = parser.parse_args(argv)

    if args.installer is not None:
        config.set_default_installer(EnvironmentType[args.installer])
        config_file.save()
        sys.exit(0)

    parser.print_usage()
    sys.exit(1)


def auth(prog: str, argv: list[str], use_keyring_if_available: bool) -> NoReturn:
    config = TomlConfigFile(DEFAULT_CONFIG_PATH)
    auth = AuthModel(config, DEFAULT_CONFIG_PATH, use_keyring_if_available=use_keyring_if_available)
    parser = _get_auth_argument_parser(prog)
    args = AuthOptions.collect(parser.parse_args(argv))

    if args.host and (":" in args.host or "/" in args.host):
        parser.error(f"invalid host name: {args.host}")
    if args.password and args.password_stdin:
        parser.error("cannot use -p,--password and --password-stdin concurrently")

    if args.remove:
        if args.list or args.username or args.password or args.password_stdin:
            parser.error("incompatible arguments")
        if not args.host:
            parser.error("missing argument `host`")
        auth.delete_credentials(args.host)
        config.save()
    elif args.list:
        if args.remove or args.host or args.username or args.password or args.password_stdin:
            parser.error("incompatible arguments")
        table = AsciiTable()
        table.headers = ["Host", "Username", "Password", "Auth check"]
        for host, username, password in auth.list_credentials():
            # Auth check
            check_result = auth_check(auth, args, host, username, password)

            table.rows.append((host, username, password if args.no_mask else "[MASKED]", check_result))
        if table.rows:
            table.print()
    elif args.username:
        if args.password_stdin:
            password = sys.stdin.readline().strip()
            if not password:
                parser.error("no password provided via stdin")
        elif args.password:
            password = args.password
        else:
            password = getpass.getpass(f"Password for {args.host}: ")
        auth.set_credentials(args.host, args.username, password)
        config.save()
    else:
        parser.print_usage()
        sys.exit(1)

    sys.exit(0)


def auth_check(auth: AuthModel, args: AuthOptions, host: str, username: str, password: str) -> str:
    check_result = "[SKIPPED]"  # Default

    if not args.no_check:
        # Check the credential now, aiming to return either OK or FAILED, and print warnings as needed

        credential_result = auth.check_credential(host, username, password)
        if credential_result:
            check_result = "[OK]" if credential_result.auth_check_result else "[FAILED]"

            # If there are any hints, output them to the logger as a warning
            if credential_result.hint:
                logger.warning(host + ": " + credential_result.hint)

            # If verbose, also display the CURL command that people can use plus the first part of the response
            if args.verbose:
                logger.info("Checking auth for host %s with command: %s", host, credential_result.curl_command)
                logger.info(
                    "First 10 lines of response (limited to 1000 chars): %s",
                    ("\n".join(credential_result.raw_result.split("\n")[0:10])[0:1000]),
                )

    return check_result


def list_pythons(prog: str, argv: list[str]) -> NoReturn:
    from kraken.common import findpython

    if argv:
        logger.error(f"{prog}: unexpected arguments")
        sys.exit(1)

    interpreters = findpython.evaluate_candidates(findpython.get_candidates(), findpython.InterpreterVersionCache())
    findpython.print_interpreters(interpreters)
    sys.exit(0)


def _print_env_status(manager: BuildEnvManager, project: Project) -> None:
    """Print the status of the environent as a nicely formatted table."""

    hash_algorithm = manager.get_hash_algorithm()

    table = AsciiTable()
    table.headers = ["Key", "Source", "Value"]
    table.rows.append(("Requirements", str(project.requirements_path), project.requirements.to_hash(hash_algorithm)))
    if project.lockfile:
        table.rows.append(("Lockfile", str(project.lockfile_path), "-"))
        table.rows.append(("  Requirements hash", "", project.lockfile.requirements.to_hash(hash_algorithm)))
        table.rows.append(("  Pinned hash", "", project.lockfile.to_pinned_requirement_spec().to_hash(hash_algorithm)))
    else:
        table.rows.append(("Lockfile", str(project.lockfile_path), "n/a"))
    if manager.exists():
        metadata = manager.get_metadata()
        environment = manager.get_environment()
        table.rows.append(("Environment", str(environment.get_path()), environment.get_type().name))
        table.rows.append(("  Metadata", str(manager.get_metadata_file()), "-"))
        table.rows.append(("    Created at", "", datetime_to_iso8601(metadata.created_at)))
        table.rows.append(("    Requirements hash", "", metadata.requirements_hash))
    else:
        table.rows.append(("Environment", str(manager.get_environment().get_path()), "n/a"))
    table.print()


def _ensure_installed(
    manager: BuildEnvManager,
    project: Project,
    reinstall: bool,
    upgrade: bool,
    env_type: EnvironmentType | None = None,
) -> None:
    exists = manager.exists()
    install = reinstall or upgrade or not exists

    operation: str
    reason: str | None = None
    allow_incremental = True

    if not exists:
        env_type = env_type or env_type or manager.get_environment().get_type()
        operation = "Initializing"
    elif upgrade:
        operation = "Upgrading"
    elif reinstall:
        operation = "Reinstalling"
    else:
        operation = "Reusing"

    current_type = manager.get_environment().get_type()
    if env_type is not None:
        type_changed = exists and env_type != current_type
        if not install and type_changed:
            install = True
            manager.remove()
            operation = "Re-initializing"
            reason = f"type changed from {current_type.name} to {env_type.name}"
            allow_incremental = False
        elif install and type_changed:
            reason = f"type changed from {current_type.name} to {env_type.name}"
            allow_incremental = False

    if not install and exists:
        metadata = manager.get_metadata()
        if project.lockfile and metadata.requirements_hash != project.lockfile.to_pinned_requirement_spec().to_hash(
            metadata.hash_algorithm
        ):
            install = True
            operation = "Re-initializing"
            reason = "outdated compared to lockfile"
        if not project.lockfile and metadata.requirements_hash != project.requirements.to_hash(metadata.hash_algorithm):
            install = True
            operation = "Re-initializing"
            reason = "outdated compared to requirements"

    if install:
        if not project.lockfile or upgrade:
            source_name = "requirements"
            source = project.requirements
            source_file = project.requirements_path
            transitive = True
        else:
            source_name = "lock file"
            source = project.lockfile.to_pinned_requirement_spec()
            source_file = project.lockfile_path
            transitive = False

        env_type = env_type or manager.get_environment().get_type()
        logger.info(
            "%s build environment from %s (%s)%s using installer %s.",
            operation,
            source_name,
            os.path.relpath(source_file),
            f" ({reason})" if reason else "",
            env_type.name,
        )

        tstart = time.perf_counter()
        manager.install(source, env_type, transitive, allow_incremental)
        duration = time.perf_counter() - tstart
        logger.info("Operation complete after %.3fs.", duration)

    else:
        logger.info("%s build environment of type %s", operation, current_type.name)


class Project(NamedTuple):
    directory: Path
    requirements_path: Path
    requirements: RequirementSpec
    lockfile_path: Path
    lockfile: Lockfile | None


def load_project(directory: Path, outdated_check: bool = True) -> Project:
    """
    This method loads the details about the current Kraken project from the current working directory
    and returns it. The project information includes the requirements for the project as well as the
    parsed lockfile, if present.

    :param directory: The directory for which to load the build project details for.
    :param outdated_check: If enabled, performs a check to see if the requirements that the lockfile was
        generated with is outdated compared to the project requirements.
    """

    project_info = GitAwareProjectFinder.default().find_project(directory)
    if not project_info:
        logger.error("no buildscript")
        sys.exit(1)
    script, runner = project_info

    # Load requirement spec from build script.
    logger.debug('Loading requirements from "%s" (runner: %s)', script, runner)

    # Extract the metadata provided by the buildscript() function call at the top of the build script.
    if not runner.has_buildscript_call(script):
        metadata = BuildscriptMetadata(requirements=["kraken-core"])
        logger.error(
            "Kraken build scripts must call the `buildscript()` function to be compatible with Kraken wrapper. "
            "Please add something like this at the top of your build script:\n\n%s\n"
            % indent(runner.get_buildscript_call_recommendation(metadata), "    "),
        )
        sys.exit(1)

    with BuildscriptMetadata.capture() as future:
        runner.execute_script(script, {})
    assert future.done()
    requirements = RequirementSpec.from_metadata(future.result())

    # Load lockfile if it exists.
    lockfile_path = script.with_suffix(".lock")
    if lockfile_path.is_file():
        logger.debug('loading lockfile from "%s"', lockfile_path)
        lockfile = Lockfile.from_path(lockfile_path)
        if outdated_check and lockfile and lockfile.requirements != requirements:
            logger.warning(
                'Lock file "%s" is outdated compared to requirements in "%s". Consider updating the lock file with '
                '"krakenw --upgrade lock"',
                os.path.relpath(lockfile_path),
                os.path.relpath(script),
            )
    else:
        lockfile = None

    return Project(script.parent, script, requirements, lockfile_path, lockfile)


@exit_on_known_exceptions(BuildEnvError, exit_code=2)
def main(krakenw_args: list[str] | None = None) -> NoReturn:
    parser = _get_argument_parser()
    args = parser.parse_args(args=krakenw_args)
    logging_options = LoggingOptions.collect(args)
    logging_options.init_logging()
    env_options = EnvOptions.collect(args)

    if not args.cmd and not env_options.any():
        parser.print_usage()
        sys.exit(0)

    # When we delegate to the Kraken CLI, we want to make sure it can detect that it has been invoked from the wrapper.
    os.environ["KRAKENW"] = "1"

    # Convert the arguments we defined in the argument parser to the actual subcommand that we want
    # delegated.
    cmd: str | None = args.cmd[0] if args.cmd else None
    argv: list[str] = args.cmd[1:] + args.args

    if cmd in ("a", "auth"):
        # The `auth` comand does not require any current project information, it can be used globally.
        auth(f"{parser.prog} auth", argv, use_keyring_if_available=not env_options.no_keyring)

    if cmd in ("config",):
        config_main(f"{parser.prog} config", argv)

    if cmd in ("list-pythons",):
        list_pythons(f"{parser.prog} list-pythons", argv)

    # The project details and build environment manager are relevant for any command that we are delegating.
    # This includes the built-in `lock` command.
    config_file = TomlConfigFile(DEFAULT_CONFIG_PATH)
    config = ConfigModel(config_file, DEFAULT_CONFIG_PATH)
    project = load_project(Path.cwd(), outdated_check=not env_options.upgrade)
    manager = BuildEnvManager(
        project.directory,
        project.directory / BUILDENV_PATH,
        AuthModel(config_file, DEFAULT_CONFIG_PATH, use_keyring_if_available=not env_options.no_keyring),
        incremental=env_options.incremental,
        show_install_logs=env_options.show_install_logs,
        default_type=config.get_default_installer(),
    )

    # Execute environment operations before delegating the command.

    is_lock_command = cmd in ("lock", "l")

    if env_options.status:
        if cmd or argv:
            logger.error("--status option must be used alone")
            sys.exit(1)
        _print_env_status(manager, project)
        sys.exit(0)

    if env_options.uninstall:
        if cmd or argv:
            logger.error("--uninstall option must be used alone")
            sys.exit(1)
        manager.remove()
        sys.exit(0)
    if env_options.any() or not is_lock_command:
        _ensure_installed(
            manager,
            project,
            env_options.reinstall,
            env_options.upgrade,
            env_options.use,
        )

    if cmd is None:
        assert not argv
        sys.exit(0)

    elif cmd in ("l", "lock"):
        lock(f"{parser.prog} lock", argv, manager, project)

    else:
        if project.directory.absolute() != Path.cwd():
            argv += ["-p", str(project.directory)]
        command = [cmd, *argv]
        logger.info("$ %s", " ".join(map(shlex.quote, ["kraken"] + command)))
        environment = manager.get_environment()
        environment.dispatch_to_kraken_cli(command)


if __name__ == "__main__":
    main()
